#
#-----------------------------------------------------------------------
# Identifier for configuration management tool TrueChange
# @(#) xml-3gpp/parsersrc/XML_3GPP_Handler.pm xml-3gpp_r3.2.1.2:cset.002139:60:60@(#)
#-----------------------------------------------------------------------
#
#   Entry Points:   New, load_handler.
#
#   XML_Parser callbacks : start_element, end_element, characters, end_document.
#
#   This module contains the callback routines required for SAX based xml 
#   parsing using the Expat XML::Parser 3rd party XML parsing tool. Each of these 
#   callbacks contain the code to perform the XML to PIF mapping. 
#
#   This module also contains the code to create an object from the module
#   and return its reference and also to configure the module with precon-
#   figured data structures. 
#
#   (c) Copyright IBM Corporation 2007. All rights reserved. 
#
package XML_3GPP_Handler;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

use AudLog;
use DDM_MANIP;
use File::Basename;

require Exporter;
require AutoLoader;

@ISA = qw(Exporter AutoLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT = qw();
$VERSION = '0.01';

my %MONTH_NAMES = ( '01' => "Jan", '02' => "Feb", '03' => "Mar",
                    '04' => "Apr", '05' => "May", '06' => "Jun",
                    '07' => "Jul", '08' => "Aug", '09' => "Sep",
                    '10' => "Oct", '11' => "Nov", '12' => "Dec" );


################################################################################
# Subroutine name:  New()
#
# Description:      Object initialisation routine
#
# Arguments:        None
#
# Returns:          Object reference
#
sub New {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};

    # These variables should be set to true of false
    # depending on whether or not the desired type of output file
    # is wanted
    bless ($self, $class);
}

################################################################################
# Subroutine name:  load_handler()
#
# Description:      Object configuration loading routine
#
# Arguments : Scalars wich contain actual and references to the following 
#             data constructs
#
#          headerElementNames           - elements for header data
#          uniquePifHeaderElementNames  - unique pif elements for header data
#          dataForPIFRecord             - The header data that is going to be 
#                                         copied into each individual record.
#          measureTypeElementSeq        - element name whose value contains 
#                                         counter names. 
#          measurementValueElementName  - element name whose value contains 
#                                         element counter values.
#          measurementResultsElementSeq - element name whose value contains
#                                         actual counter values.
#          headerInfoForPif             - Header data to be used in PIF file
#                                         name.
#          externalheaderData           - external data for PIF header
#          outputBlockElementName       - Element name whose value holds O/P
#                                         block name
#          outputBlockDescription       - Allows output Block name to be 
#                                         configurable.
#          instanceObject               - Current measurement object instance 
#                                         ID (moid) element name.
#          suspectFlagElementName       - Element name containing suspect flag
#                                         value.
#          listDelimiter                - Delimiter used for separating counter
#                                         names/values within a single element tag. 
#          altCounterNullValue          - Alternative to null in PIF output.
#          altOutputBlockNames          - Alternative block names.
#          altCounterNames              - Alternative counter names.
#          uniquePIFBlockElementName    - Element which determines XML to be 
#                                         separated out into separate PIF files
#          maxCounterNameSize           - Size which the counter name can not be
#                                         greater than
#          int_dir                      - Intermediate directory path. 
#          ddmManipRules                - The configuration of DDM counters to 
#                                         be manipulated.
#          enablePIFCounterInFilename   - Boolean to determine whether a counter
#                                         should be included in the PIF filename
#                                         output.
#          enablePIFPerMoid             - create a single pif per moid key
#          headerRecordProcessing       - a subroutine written in the EngineConfig
#                                         for header record.  Contains functionality 
#                                         that cannot be found in the engine.
#          dataRecordProcessing         - a subroutine written in the EngineConfig 
#                                         for data block.  Extra functionality to 
#                                         the engine.
#
#
#
# Returns:          0 for success,
#                   

sub load_handler {
    my $self = shift;
    
    #Copy in arguments to object.
    ($self->{headerElementNames},
     $self->{uniquePifHeaderElementNames},
     $self->{dataForPIFRecord},
     $self->{measureTypeElementSeq},
     $self->{measurementValueElementName},
     $self->{measurementResultsElementSeq},
     $self->{headerInfoForPif},
     $self->{externalHeaderData},
     $self->{outputBlockElementName},
     $self->{outputBlockDescription},
     $self->{instanceObject},
     $self->{suspectFlagElementName},
     $self->{listDelimiter},
     $self->{altCounterNullValue},
     $self->{altOutputBlockNames},
     $self->{altCounterNames},
     $self->{uniquePIFBlockElementName},
     $self->{maxCounterNameSize},
     $self->{int_dir},
     $self->{ddmManipRules},
     $self->{enablePIFCounterInFilename},
     $self->{enablePIFPerMoid},
     $self->{headerRecordProcessing},
     $self->{dataRecordProcessing}) = @_;

    #Initialise more object data required.
    $self->{maxSizeHeaderElementNames} = scalar(keys %{$self->{uniquePifHeaderElementNames}}); 
    $self->{measurementTypes} = [];
    $self->{measurementValues} = [];
    $self->{ddmMeasurementTypes} = [];
    $self->{ddmMeasurementValues} = [];
    $self->{pifDataRecord} = {}; 
    $self->{enablePifDataRecord} = ();
    $self->{blockName} = ();
    $self->{blockNameCount} = {};
    $self->{PIF_COUNT} = 0;
    $self->{elementName} = "";
    $self->{outputPifData} = {};
    $self->{pif_counter} = 0;

    # We need to capture whether the PIF blockname is the same for all PIF's, 
    # or will change within every <mi> block. Defaults to common.
    $self->{isBlockNameCommon} = "True";
    $self->{suspectFlagIsTrue} = ();
    $self->{currentElementName} = ();

    # use for common moids per PIF
    $self->{PIF_OBJECTS} = ();
    $self->{commonPifData} = ();
    $self->{headerForPifDataRecord} = () if (!(exists($self->{headerForPifDataRecord})));

    # Initialise optional items if undefined.
    if ( ! $self->{internalHeaderData} ) {
       $self->{internalHeaderData} = {};
    }

    if ( ! $self->{altCounterNames} ) {
       $self->{altCounterNames} = [];
    }

    if ( ( $self->{externalHeaderData} ) && 
         ( $self->{internalHeaderData} ) ) {
       # Integrate external file/directory name header data into internal
       # buffers
       foreach ( keys %{$self->{externalHeaderData}} ) {
          $self->{commonInternalHeaderData}->{$_} = $self->{externalHeaderData}->{$_};
       }
    }
    return 0;
}

################################################################################
#
# Subroutine name:  start_element()
#
# Description:      SAX callback for new element. Also handles any attributes
#                   found. 
#
# Arguments:        currentElementName (scalar) - current element being parsed. 
#
# Returns:          none.
#
################################################################################

sub start_element {
   my $self = shift;
   my $ignore = shift;
   $self->{currentElementName} = shift;
   
   my @attrs = @_;
   
   # If there is no attribute, do not populate $self->{currentAttributeRef}
   if (scalar(@attrs) != 0) {
       undef $self->{currentAttributeRef};
       foreach (@attrs) {
           push @{$self->{currentAttributeRef}},($_);
       } 
   }
 
   my %ctrPair;
   my ($attrName, $attrValue);

   $self->{elementName} = $self->{currentElementName};

   # This is to clear $self->{data} and contains the characters data everytime
   # the start_element is invoked.
   undef $self->{data};

   # Enable new PIF record if mv
   if ( $self->{measurementValueElementName}->{$self->{currentElementName}}) {
      $self->{enablePifDataRecord} = 'True';
   }

   # If current element is required for header data, include its attributes 
   # also.
   if ( $self->{headerElementNames}->{ $self->{currentElementName}} || 
        $self->{uniquePifHeaderElementNames}->{ $self->{currentElementName}} ) {
      for (my $i=0; $i < scalar(@attrs); $i++) {
         $attrName = $attrs[$i];
         $attrValue = $attrs[++$i];
         %ctrPair = $self->process_attribute($attrName, $attrValue);
         while ( my($ctr,$value) = each %ctrPair ) {
            $ctr = $self->{currentElementName}.'_'.$ctr;
            $self->processAltCounterNames(\$ctr);
            if ( $self->{headerElementNames}->{ $self->{currentElementName}} ) {
               $self->{commonInternalHeaderData}->{$ctr} = $value;
            }
            else {
               $self->{internalHeaderData}->{$ctr} = $value;
            }
            # header data goes to PIF records.
            if ( $self->{dataForPIFRecord}->{$self->{currentElementName}} ) {
               $self->{headerForPifDataRecord}->{$ctr} = $value;
            }
         }
      }
   }

   if ( $self->{uniquePIFBlockElementName}->{$self->{currentElementName}}) {

      $self->{internalHeaderData} = {};
      $self->{pif_data_record_counter} = 0;

      if  ( ! $self->{blockName} ) {
         $self->{isBlockNameCommon} = 0;
      }
   }
}


################################################################################
#
# Subroutine name:  processAltCounterNames
#
# Description:      Counter Names can be altered by configure the new name in
#                   EngineConfig.pm
#                   Note the function can handle global replacements and element
#                   extractions and insertion.
#
# Arguments:        counterNameRef  - Reference to counter to be altered.
#
# Returns:          none.
#
################################################################################
sub processAltCounterNames {
   my $self = shift;
   my $counterNameRef = shift;
   my @matches;

   for (my $i=0; $i < scalar(@{$self->{altCounterNames}}); $i++) {
       if ($self->{altCounterNames}[$i][1] =~ /\$\d/o){
           if ( @matches = $$counterNameRef =~ /$self->{altCounterNames}[$i][0]/) {
               $$counterNameRef = $self->{altCounterNames}[$i][1];
               for (my $i=1; $i <= @matches; $i++){
                   $$counterNameRef =~ s/\$$i/$matches[$i-1]/;
               }
                   
               last;
           }
       }
       elsif ($$counterNameRef =~ s/$self->{altCounterNames}[$i][0]/$self->{altCounterNames}[$i][1]/g){
           last;
       }
           
   }
}

################################################################################
#
# Subroutine name:  characters()
#
# Description:      SAX callback for element characters value. Because of the 
#                   limitation of XML::Parser that will break a string of character
#                   when it reaches certain buffer limit, this subroutine will
#                   only concatenating the whole string character before being 
#                   process by process_characters.
#
# Arguments:        data (scalar) - current characters value being parsed. 
#
# Returns:          none.
#
################################################################################
sub characters {
    my $self = shift;
    my $ignore = shift;
    my $data = shift;

    $self->{data} .= $data;
}

################################################################################
#
# Subroutine name:  process_characters()
#
# Description:      This subroutine will process a string of character and action 
#                   depend on how element name is configured.  
#
# Arguments:        data (scalar) - current character value being parsed. 
#
# Returns:          none.
#
################################################################################
sub process_characters {
    my $self = shift;
    my $data = $self->{data};
    my %tmpHash;
    my ($ctr, $value);

    # Check for ~.
    $data =~ s/~/_/g;

    # Check for empty data.
    $data =~ s/""//o;
    
    # remove initial non-alphanumeric characters.
    $data =~ s/^[\n\t\f\r]*(.*)/$1/;

    # Check for equals.
    if ($data =~ /=/o ) {
        %tmpHash = $data =~ /([^=\s,]+)\s*=[\s"]*([^,\s"]*)[\s"]*/g;
    }

    if ( uc($data) eq "NULL") {
        if (  ( exists($self->{altCounterNullValue}) ) && 
              ($self->{altCounterNullValue}) ) {
            $data = $self->{altCounterNullValue};
        } 
        else {
            $data = "";
        }
    }
    
    # Format the date and time into DDMONYYYY HH:MM:SS
    if ( $data =~ /^(20\d{2})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})$/o) {
        if ($2 > 0 && $2 < 13 && $3 > 0 && $3 < 32 && $4 < 24 &&
            $5 < 60 && $6 < 60){
            # a valid date - convert
            #$data = $3.$MONTH_NAMES{$2}.$1."_".$4.":".$5.":".$6;
            $data = $3.$MONTH_NAMES{$2}.$1."_".$4.":".$5.":00";
        }
    }

    # Updated to find all Date time values 
    if ( $data =~ /^(20\d{2})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})[\.\d\+\d{4}]/o) {
        if ($2 > 0 && $2 < 13 && $3 > 0 && $3 < 32 && $4 < 24 &&
            $5 < 60 && $6 < 60){
            # a valid date - convert
            #$data = $3.$MONTH_NAMES{$2}.$1."_".$4.":".$5.":".$6;
            $data = $3.$MONTH_NAMES{$2}.$1."_".$4.":".$5.":00";
        }
    }
    
    # Block name for file and also alternative block name if configured.
    if ($self->{outputBlockElementName}->{$self->{currentElementName}}) {
        my $blk;
      
        if ( $self->{outputBlockDescription} ) {
            my (@ary, $i);
            if ( @ary = $data =~ /$self->{outputBlockDescription}/g ) {
                my $tmp = join(",", @ary);
                foreach $i (@ary) {
                    if ($i) {
                        $blk .= $i.'_';
                    } 
                    else {
                        last;
                    }
                }
                $self->{blockName} = $blk;
            } 
            elsif ( @ary = $self->{currentAttributeRef}[1] =~ /$self->{outputBlockDescription}/g ) {
                foreach $i (@ary) {
                    $i =~ s/^\s*(.+)/$1/;
                    if ($i) {
                        $blk .= $i.'_';
                    } 
                    else {
                        last;
                    }
                }
                $self->{blockName} = $blk;
                $self->{blockName} =~ tr/ /_/;
            }
            else {
                LogMess("Error - output block description match failed ",2);
                $self->{blockName} = 'BLOCKNAME';
            }
        } 
        else {
            $self->{blockName} = $data;
        }
        $self->{blockName} =~ s/^(.+)_$/$1/;
        RE:
        foreach my $re ( keys %{$self->{altOutputBlockNames}} ) {
            if ( $self->{blockName} =~ /$re/ ) {
                $self->{blockName} = $self->{altOutputBlockNames}->{$re};
                last RE;
            }
        }
    }

    # Process char and current element name to PIF header if configured.  
    if ( $self->{headerElementNames}->{ $self->{currentElementName}} ||
         $self->{uniquePifHeaderElementNames}->{ $self->{currentElementName}} ) {
        if ( $data =~  /^(\d{2}\w{3}\d{4})\_(\d{2}.+)$/o) {
            my $dateCtr = $self->{currentElementName}.'_'.'Date';
            my $timeCtr = $self->{currentElementName}.'_'.'Time';
            if ($self->{enablePIFPerMoid} ) {
                $self->{commonPifData}->{$dateCtr} = $1;
                $self->{commonPifData}->{$timeCtr} = $2;
            }
            else {
                if ( $self->{headerElementNames}->{ $self->{currentElementName}} ) {
                    $self->{commonInternalHeaderData}->{$dateCtr} = $1;   
                    $self->{commonInternalHeaderData}->{$timeCtr} = $2;
                }
                else {
                    $self->{internalHeaderData}->{$dateCtr} = $1;   
                    $self->{internalHeaderData}->{$timeCtr} = $2;
                }
            }
        }

        # Rule - if '=' is present in element data (this is not an attribute); 
        # the header name is obtained from the LHS of '=' and the header value 
        # from the RHS. Otherwise, the element tag is the counter name and the 
        # value is the counter value.
        if ($data =~ /=/o ) {
            while ( ($ctr,$value) = each %tmpHash ) {
                if ( $value =~ /(.*)=(.*)/o ) {
                    $ctr = $1;
                    $value = $2;
                }
                $value =~ s/(.+\s.*)/\"$1\"/;
                $ctr = $self->{currentElementName}.'_'.$ctr;
                $self->processAltCounterNames(\$ctr);
                if ($self->{enablePIFPerMoid} ) {
                    $self->{commonPifData}->{$ctr} = $value;
                }
                else {
                    if ( $self->{headerElementNames}->{ $self->{currentElementName}} ) {
                        $self->{commonInternalHeaderData}->{$ctr} = $value;
                    }
                    else {
                        $self->{internalHeaderData}->{$ctr} = $value;
                        if ($self->{commonInternalHeaderData}->{$ctr}) {
                            $self->{commonInternalHeaderData}->{$ctr} = $value;
                        }
                    }
                }
                # header data for PIF record
                if ( $self->{dataForPIFRecord}->{$self->{currentElementName}} ) {
                    $self->{headerForPifDataRecord}->{$ctr} = $value;
                }
            }
        } 
        else {
            $ctr = $self->{currentElementName};
            $data =~ s/(.+\s.*)/\"$1\"/;
            $self->processAltCounterNames(\$ctr);
            if ($self->{enablePIFPerMoid} and !($self->{headerElementNames}->{$self->{currentElementName}})
                and !$self->{uniquePifHeaderElementNames}->{ $self->{currentElementName}} ) {
                $self->{commonPifData}->{$ctr} = $data;
            }
            else {
                if ( $self->{headerElementNames}->{ $self->{currentElementName}} ) {
                    $self->{commonInternalHeaderData}->{$ctr} = $data;
                }
                else {
                    $self->{internalHeaderData}->{$ctr} = $data;
                    if ($self->{commonInternalHeaderData}->{$ctr}) {
                        $self->{commonInternalHeaderData}->{$ctr} = $data;
                    }
                }
            }
            # header data for PIF record
            if ( $self->{dataForPIFRecord}->{$self->{currentElementName}} ) {
                $self->{headerForPifDataRecord}->{$ctr} = $data;
            }
        }
        
    }
    
    if ( $self->{measureTypeElementSeq}->{ $self->{currentElementName}} ) {
        if ( $self->{listDelimiter} ) {
            # attempt to split data using the provided delimiter, 
            # if successful the counter name is in list.
            my @counterNames = split($self->{listDelimiter},$data);
            foreach my $counterName (@counterNames) {
                $self->processAltCounterNames(\$counterName);
                $counterName =~ s/^(.{$self->{maxCounterNameSize}}).*/$1/;
                $counterName =~ s/(.+\s.*)/\"$1\"/;
                push @{$self->{measurementTypes}}, ($counterName);
            }
        }
        else {
            $self->processAltCounterNames(\$data);

            # Truncate counter names by configured amount
            $data =~ s/^(.{$self->{maxCounterNameSize}}).*/$1/;
            $data =~ s/(.+\s.*)/\"$1\"/;
            push @{$self->{measurementTypes}}, ($data);
        }
    }
   
    if ( $self->{enablePifDataRecord} ) {
        
        # Process moid for each record. 
        if ($self->{instanceObject}->{$self->{currentElementName}}) {
            if ($data =~ /=/o ) {
                while ( ($ctr,$value) = each %tmpHash ) {
                    $value =~ s/(.+\s.*)/\"$1\"/;
                    $ctr = $self->{currentElementName}.'_'.$ctr;
                    $self->processAltCounterNames(\$ctr);
                    $self->{pifDataRecord}->{$ctr} = $value;
                }
            } 
            elsif ( $self->{currentAttributeRef}[1] =~ /=/o ) {
                %tmpHash = $self->process_attribute($self->{currentAttributeRef}[0], 
                                             $self->{currentAttributeRef}[1]);
                while ( ($ctr,$value) = each %tmpHash ) {
                    $ctr = $self->{currentElementName}.'_'.$ctr;
                    $self->processAltCounterNames(\$ctr);
                    $self->{pifDataRecord}->{$ctr} = $value;
                }
            }
            else {
                $ctr = $self->{currentElementName};
                $self->processAltCounterNames(\$ctr);
                $data =~ s/(.+\s.*)/\"$1\"/;
                $self->{pifDataRecord}->{$ctr} = $data;
            }
        }
        
        # Collect counter values for new record.
        if ($self->{measurementResultsElementSeq}->{$self->{currentElementName}}) {
            my $measType = @{$self->{measurementTypes}}[scalar @{$self->{measurementValues}}];
            
            if (exists($self->{ddmManipRules}{COUNTERS_TO_MANIP}{$measType})){
                DDM_MANIP::ddm_manip($self,$measType,$data);
                # put a dummy value on the measurementValues for the measurementType
                push @{$self->{measurementValues}}, 'DUMMY';
            }
            elsif ( $self->{listDelimiter} ) {
                # Try to split the data with provided delimiter,
                #    if successful the values are in list 
                my @counterValues = split( $self->{listDelimiter}, $data);
                push @{$self->{measurementValues}}, (@counterValues);
            }
            else {
                $data =~ s/(.+\s.*)/\"$1\"/;
                push @{$self->{measurementValues}}, ($data);
            }
        }
    }
    
    if ($self->{suspectFlagElementName}->{$self->{currentElementName}}) {
        
        if ( uc($data) eq "TRUE" ) {
            LogMess("Warning : Suspect flag $data; Current measurement values record is unreliable ",3);
            LogMess("Warning : record will be omitted ",3);
            $self->{suspectFlagIsTrue} = "True";
        } 
        else {
            $ctr = $self->{currentElementName};
            $self->processAltCounterNames(\$ctr);
            $self->{pifDataRecord}->{$ctr} = $data;
        }
    }
    
    %tmpHash = ();

}

################################################################################
#
# Subroutine name:  end_element()
#
# Description:      SAX callback for element end tag. 
#
# Arguments:        currentElementName (scalar) - current end element being 
#                   parsed. 
#
# Returns:          none.
#
################################################################################

sub end_element {
   my $self = shift;
   my $ignore = shift;
   $self->{currentElementName} = shift;

   my (%pifData, $pif_key, $hash_code);

   # After all the string characters have been gathered, now process the characters
   if ( ($self->{measurementResultsElementSeq}->{$self->{currentElementName}}) ||  # Counter values
        ($self->{measureTypeElementSeq}->{ $self->{currentElementName}}) ||        # Counter names
        ($self->{uniquePifHeaderElementNames}->{ $self->{currentElementName}}) ||
        ($self->{outputBlockElementName}->{$self->{currentElementName}}) ||
        ($self->{instanceObject}->{$self->{currentElementName}}) ||
        ($self->{suspectFlagElementName}->{$self->{currentElementName}}) || 
        ($self->{headerElementNames}->{ $self->{currentElementName}}) ) { 
       $self->process_characters();
       undef $self->{data};
   }

   if ( $self->{measurementValueElementName}->{$self->{currentElementName}}) {
      $self->{enablePifDataRecord} = 0;

      unless($self->{suspectFlagIsTrue}) {
         @pifData{@{$self->{measurementTypes}}} = @{$self->{measurementValues}};
         @pifData{keys %{$self->{pifDataRecord}}} = values %{$self->{pifDataRecord}};
         
         # delete ddm measurement types which have derived values
         foreach (keys %{$self->{ddmManipRules}{COUNTERS_TO_MANIP}}){
             delete $pifData{$_};
         }
         # and add any ddm measurement types derived.
         @pifData{@{$self->{ddmMeasurementTypes}}} = @{$self->{ddmMeasurementValues}};

         if ( exists($self->{headerForPifDataRecord}) ) {
             @pifData{keys %{$self->{headerForPifDataRecord}}} = values %{$self->{headerForPifDataRecord}};
         }

         if ($self->{enablePIFPerMoid} ) {
             @pifData{keys %{$self->{commonPifData}}} = values %{$self->{commonPifData}};
         }

         # Populate the outputPifData
         $self->{outputPifData}->{$self->{pif_counter}}->{blockName} = $self->{blockName};
         @{$self->{outputPifData}->{$self->{pif_counter}}->{internalHeaderData}}{keys %{$self->{internalHeaderData}}} = values %{$self->{internalHeaderData}};
         @{$self->{outputPifData}->{$self->{pif_counter}}->{$self->{pif_data_record_counter}}->{pifData}}{keys
%pifData} = values %pifData;
         $self->{pif_data_record_counter}++;
      }

      %pifData = ();
      $self->{pifDataRecord} = {};
      $self->{suspectFlagIsTrue} = ();
      $self->{measurementValues} = [];
      $self->{ddmMeasurementValues} = [];
      $self->{ddmMeasurementTypes} = [];
   }

   if ( $self->{uniquePIFBlockElementName}->{$self->{currentElementName}} ) {
       $self->{measurementTypes} = [];
       $self->{ddmMeasurementTypes} = [];
       $self->{pif_counter}++;
   }
   $self->{currentElementName} = ();
}

################################################################################
#
# Subroutine name:  end_document()
#
# Description:      SAX callback for end of document
#
# Arguments:        currentElementName (scalar) - final end element being 
#                   parsed. 
#
# Returns:          none.
#
################################################################################

sub end_document {
   my $self = shift;
   my $ignore = shift;
   my @pifKeys = [];
   my $pif_key = ();
   my $hash_code = ();
   my $i = 0;
   my $j;
   
   # Start writing PIF
   
   if ($self->{enablePIFPerMoid}) {
       # If PIF per moid, write PIF per record
   
       foreach $i (keys %{$self->{outputPifData}}) {
           foreach (keys %{$self->{commonInternalHeaderData}}) {
               $self->{outputPifData}->{$i}->{internalHeaderData}->{$_} = $self->{commonInternalHeaderData}->{$_};
           }
   
           @{$self->{internalHeaderData}}{keys %{$self->{outputPifData}->{$i}->{internalHeaderData}}} = values %{$self->{outputPifData}->{$i}->{internalHeaderData}};
   
           $self->{blockName} = $self->{outputPifData}->{$i}->{blockName};

           # execute external subroutine for header record if configured (made redundant by HEADER_DATA_RECORD_PROCESSING)
	   if ( (! exists($self->{HEADER_DATA_RECORD_PROCESSING})) && $self->{headerRecordProcessing} ) {
               $self->invoke_configured_record_processing(
                   $self->{headerRecordProcessing},
                   $self->{outputPifData}->{$i}->{internalHeaderData});
           }

           # Generate PIF for each moid, write PIF data and close PIF
           my $outputPifCounterRef = $self->{outputPifData}->{$i};
NEXT_REC:  foreach $j (keys %{$outputPifCounterRef}) {
               next NEXT_REC if ( $j eq "blockName" );
               next NEXT_REC if ( $j eq "internalHeaderData" );
               
               # execute external subroutine for header and data record if configured
	       if ($self->{HEADER_DATA_RECORD_PROCESSING}) {
		   my $block_name = $self->{blockName};
                   $self->invoke_header_data_record_processing(
		       \$block_name,
                       $self->{outputPifData}->{$i}->{internalHeaderData},
                       $self->{outputPifData}->{$i}->{$j}->{pifData}
	           );
	           $self->{blockName} = $block_name;
	       }
  
               ($pif_key, $hash_code) = $self->get_pif_key($self->{outputPifData}->{$i}->{$j}->{pifData});

               if ( !exists($self->{PIF_OBJECTS}->{$pif_key}) ||
                    $hash_code ne $self->{PIF_OBJECTS}->{$pif_key}->{HASH_CODE} ) {
                   if ( exists( $self->{PIF_OBJECTS}->{$self->{CURRENT_PIF_KEY}} ) ) {
                       $self->{PIF_OBJECTS}->{$self->{CURRENT_PIF_KEY}}->{PIF}->Close_PIF_Write (
                           $self->{PIF_OBJECTS}->{$self->{CURRENT_PIF_KEY}}->{CTX}
                       );
                       $self->{PIF_OBJECTS}->{$self->{CURRENT_PIF_KEY}}->{CLOSE_FLAG} = 'True';
                   }

                   if (!exists($self->{PIF_OBJECTS}->{$pif_key})){
                       ($self->{PIF_OBJECTS}->{$pif_key}->{PIF},
                        $self->{PIF_OBJECTS}->{$pif_key}->{CTX}) = $self->init_pif($pif_key);

                        $self->{PIF_OBJECTS}->{$pif_key}->{PIF}->PIF_Write($self->{PIF_OBJECTS}->{$pif_key}->{CTX},
                                                                           'HEADER',
                                                                           $self->{outputPifData}->{$i}->{internalHeaderData});
                        delete $self->{PIF_OBJECTS}->{$pif_key}->{CLOSE_FLAG};
                   }
                   elsif ($hash_code ne $self->{PIF_OBJECTS}->{$pif_key}->{HASH_CODE}){
                       my $unique_pif = $pif_key . $self->{PIF_COUNT}++ . "-#-";
                       ($self->{PIF_OBJECTS}->{$pif_key}->{PIF},
                        $self->{PIF_OBJECTS}->{$pif_key}->{CTX}) = $self->init_pif($unique_pif);

                       $self->{PIF_OBJECTS}->{$pif_key}->{PIF}->PIF_Write($self->{PIF_OBJECTS}->{$pif_key}->{CTX},
                                                                          'HEADER',
                                                                          $self->{outputPifData}->{$i}->{internalHeaderData});
                       delete $self->{PIF_OBJECTS}->{$pif_key}->{CLOSE_FLAG};
                   }
                   $self->{CURRENT_PIF_KEY} = $pif_key;
               }
               $self->{PIF_OBJECTS}->{$pif_key}->{HASH_CODE} = $hash_code;

               if (defined($self->{outputPifData}->{$i}->{$j}->{pifData}) && $self->{PIF_OBJECTS}->{$pif_key}->{PIF}) {
                   my $ctrCounts = 0;
                   foreach(keys %{$self->{outputPifData}->{$i}->{$j}->{pifData}}) {
                       $ctrCounts++ if ($_ ne '');
                   }
                   $self->{outputPifData}->{$i}->{$j}->{pifData}{'Record_measurement_count'} = $ctrCounts;

                   # execute external subroutine for data block if configured (made redundant by HEADER_DATA_RECORD_PROCESSING)
	           if ( (! exists($self->{HEADER_DATA_RECORD_PROCESSING})) && $self->{dataRecordProcessing} ) {
                       $self->invoke_configured_record_processing ($self->{dataRecordProcessing},
                                                                   $self->{outputPifData}->{$i}->{$j}->{pifData});
                   }

                   if ( $self->{blockName} ) {
                       $self->{PIF_OBJECTS}->{$pif_key}->{PIF}->PIF_Write($self->{PIF_OBJECTS}->{$pif_key}->{CTX},
                                                                          $self->{blockName},
                                                                          $self->{outputPifData}->{$i}->{$j}->{pifData});
                   }
                   else {
                       LogMess("Warning - no blockname found - default to \'BLOCKNAME\'",3);
                       LogMess("Warning - check XML_OUTPUT_BLOCK_ELEMENT_NAME config entry",3);
                       $self->{PIF_OBJECTS}->{$pif_key}->{PIF}->PIF_Write($self->{PIF_OBJECTS}->{$pif_key}->{CTX},
                                                                          'BLOCKNAME',
                                                                          $self->{outputPifData}->{$i}->{$j}->{pifData});
                   }
               }
           }
       }
       # if we're creating a PIF per moid, there may be many PIFs open
       # to be closed.
       @pifKeys = keys %{$self->{PIF_OBJECTS}};
       if ( scalar(@pifKeys) > 0 ) {
           foreach (@pifKeys){
               if ( $self->{PIF_OBJECTS}->{$_}->{PIF} &&
                    !exists($self->{PIF_OBJECTS}->{$_}->{CLOSE_FLAG})) {
                   $self->{PIF_OBJECTS}->{$_}->{PIF}->Close_PIF_Write($self->{PIF_OBJECTS}->{$_}->{CTX});
               }
           }
       }

   }
   else {
       # If not PIF per moid, write PIF per mi
   
       foreach $i (keys %{$self->{outputPifData}}) {
           foreach (keys %{$self->{commonInternalHeaderData}}) {
               $self->{outputPifData}->{$i}->{internalHeaderData}->{$_} = $self->{commonInternalHeaderData}->{$_};
           }

           @{$self->{internalHeaderData}}{keys %{$self->{outputPifData}->{$i}->{internalHeaderData}}} = values %{$self->{outputPifData}->{$i}->{internalHeaderData}};

           $self->{blockName} = $self->{outputPifData}->{$i}->{blockName};

           # execute external subroutine for header record if configured (made redudant by HEADER_DATA_RECORD_PROCESSING)
	   if ( (! exists($self->{HEADER_DATA_RECORD_PROCESSING})) && $self->{headerRecordProcessing} ) {
               $self->invoke_configured_record_processing (
                   $self->{headerRecordProcessing},
                   $self->{outputPifData}->{$i}->{internalHeaderData});
           }

           # Generate PIF for each mi, write PIF data and close PIF  
           my $outputPifCounterRef = $self->{outputPifData}->{$i};
NEXT_REC:  foreach $j (keys %{$outputPifCounterRef}) {
               next NEXT_REC if ( $j eq "blockName" );
               next NEXT_REC if ( $j eq "internalHeaderData" );

               # execute external subroutine for header and data record if configured
	       if ($self->{HEADER_DATA_RECORD_PROCESSING}) {
		   my $block_name = $self->{blockName};
                   $self->invoke_header_data_record_processing(
		       \$block_name,
                       $self->{outputPifData}->{$i}->{internalHeaderData},
                       $self->{outputPifData}->{$i}->{$j}->{pifData}
	           );
	           $self->{blockName} = $block_name;
	       }
  
               ($pif_key, $hash_code) = $self->get_pif_key($self->{outputPifData}->{$i}->{$j}->{pifData});

               if (!exists($self->{PIF_OBJECTS}->{notPifPerMoid}) ) {
                   ($self->{PIF_OBJECTS}->{notPifPerMoid}->{PIF},
                    $self->{PIF_OBJECTS}->{notPifPerMoid}->{CTX}) = $self->init_pif($pif_key,0);
                   $self->{PIF_OBJECTS}->{notPifPerMoid}->{PIF}->PIF_Write($self->{PIF_OBJECTS}->{notPifPerMoid}->{CTX},
                                                                           'HEADER',
                                                                           $self->{outputPifData}->{$i}->{internalHeaderData});
               }
               $pif_key = 'notPifPerMoid';

               if (defined($self->{outputPifData}->{$i}->{$j}->{pifData}) &&
                   $self->{PIF_OBJECTS}->{$pif_key}->{PIF}) {
                   my $ctrCounts = 0;
                   foreach(keys %{$self->{outputPifData}->{$i}->{$j}->{pifData}}) {
                       $ctrCounts++ if ($_ ne '');
                   }
                   $self->{outputPifData}->{$i}->{$j}->{pifData}{'Record_measurement_count'} = $ctrCounts;

                   # execute external subroutine for data block if configured (made redudant by headerDataRecrodProcessing)
	           if ( (! exists($self->{HEADER_DATA_RECORD_PROCESSING})) && $self->{dataRecordProcessing} ) {
                       $self->invoke_configured_record_processing ($self->{dataRecordProcessing},
                       $self->{outputPifData}->{$i}->{$j}->{pifData});
                   }

                   if ( $self->{blockName} ) {
                       $self->{PIF_OBJECTS}->{$pif_key}->{PIF}->PIF_Write($self->{PIF_OBJECTS}->{$pif_key}->{CTX},
                                                                          $self->{blockName},
                                                                          $self->{outputPifData}->{$i}->{$j}->{pifData}
                                                                          );
                   }
               }
           }
           $self->{PIF_OBJECTS}->{$pif_key}->{PIF}->Close_PIF_Write($self->{PIF_OBJECTS}->{$pif_key}->{CTX});
           delete $self->{PIF_OBJECTS}->{$pif_key};
       }
   } 
}


################################################################################
#
# Subroutine name:  init_pif()
#
# Description:      Initialise a PIF file and return the object id that is 
#                   dealing with it.
#
# Arguments:        none 
#                  
# Returns:          PIF object ID reference
#
################################################################################
sub init_pif {
    my $self = shift;
    my $pif_key = shift;
    my ($objid, $objctx, $d, $filename);
    
    $filename = $pif_key;
    $filename .= 'I';
    $filename =~ s/\//_/g;
    $filename =~ s/\s+/_/g;
    
    # Opening the PIF and writting the header information out
    ($objid, $objctx) = GenUtils->get_pif_object();
    $objid->Open_PIF_Write($objctx, $self->{int_dir}.'/'.$filename.'.pif');
    return ($objid, $objctx); 
}

################################################################################
#
# Subroutine name: get_pif_key
#
# Description: creates the PIF filename key based on the configuration
#              
#
# Arguments: none
#
# Returns: a unique which can be used in the PIF name, and to identify the PIF
#
################################################################################
sub get_pif_key(){
    my $self = shift;
    my $pifData = shift;
    
    my ($pif_key,$hash_code);

#    my %tmp_hash = reverse %{$self->{headerInfoForPif}};
#    foreach (sort keys %tmp_hash){
#        if ( $self->{internalHeaderData}->{$tmp_hash{$_}} ) {
#            $pif_key .= $self->{internalHeaderData}->{$tmp_hash{$_}} . '-#-';
#        }
    foreach ( keys %{$self->{headerInfoForPif}} ) {
        if ( $self->{internalHeaderData}->{$_} ) {
            $pif_key .= $self->{internalHeaderData}->{$_} . '-#-';
        }
    }
    
    
    if ($self->{blockName}) {
        $pif_key = $self->{blockName} . '-#-' . $pif_key;
    }

    if ($self->{enablePIFPerMoid}){
        # extract moids from pifData to create unique key
        foreach (keys %$pifData){
            if ($_ =~ /moid_(.*)$/o){
                $pif_key .= $1 . "-#-";
            }
            
        }
    }
    elsif ( $self->{enablePIFCounterInFilename} ) {
        $self->{blockNameCount}->{$self->{blockName}} = 1 
                  if (!($self->{blockNameCount}->{$self->{blockName}}));
        $pif_key .= $self->{blockNameCount}->{$self->{blockName}}.'-#-'; 
        $self->{blockNameCount}->{$self->{blockName}}++;
    } 

    $hash_code = join("-#-", sort keys %$pifData);
    return ($pif_key,$hash_code);
}

################################################################################
#
# Subroutine name:  process_attribute
#
# Description:      Breaks attribute values into counter/value pairs as below:
# - LHS of '=' is counter name, and RHS of '=' is counter value.
# - Further breakdown of counter value if it consists '=' in it.
# - Preceding spaces in counter name should be removed and others are replaced with '_'.
# - For further breakdown LHS value should be appended to the original LHS value.
# - If counter value ended with numeric value, non numeric value will be appended 
#   to LHS value with trailing '_ID'.  Counter value will take the number.
# - If RHS value consist timestamp format, it breaks into date, time, GMT difference.
# - If RHS value in /^PT(\d{4})S$/, counter value is $1.
# - Counter value is wrapped with double quotes if there is space in it.
#
# Arguments:       $atribName - the attribute name
#                  $atribVal - the attribute value
#
# Returns:         %ctrPair - the breaks up counter name/value pairs.
#
################################################################################

sub process_attribute {

   my ($self, $atribName, $atribVal) = @_;
   my %ctrPair;
   my ($ctr,$value,$tmpString);
   my %tmpHash = $atribVal =~ /([\w_\-\.\/\\]+)=\"?([\w\s_\-\.\/\\:=]*)\"?/g;

   if ( %tmpHash ) {
      while ( ($ctr,$value) = each %tmpHash ) {
         if ( $value =~ /.+-\d+.*/o ) {       # multiple instances ABC-9 DEF-2 ...
            my %tmpHash2 = $value =~ /\s*([\w_\/\\\s]+)\.*\-(\d+)/g;
            while ( my($ctr2,$value2) = each %tmpHash2 ) {
               $tmpString = $ctr.'_'.$ctr2;
               $tmpString =~ s/^\s*(.+)/$1/;
               $tmpString =~ tr/ /_/;
               $tmpString .= '_ID' if ($tmpString !~ /.*_ID$/);
               $ctrPair{$tmpString}=$value2;
            }
         }
         else {
            if ( $value =~ /(.+)=(.+)/o ) {
               $ctr .= '_'.$1;
               $value = $2;
            }
            $value =~ s/.*:\.(.*)/$1/;             #  abcdef:.(ABCDEF)
            $value =~ s/.*:([\w\s]*\.*-*\d*$)/$1/;  #  abcdef:(AB CD EF.-99) abcdef:(ADBN_LKHKD_KDH)
            if ( $value =~ /(.+)\.*-(\d+)$/o ) {
               $value = $2;
               $ctr .= '_'.$1;
               $ctr .= ' ID' if ( $ctr !~ /\sID$/ );
            }
            $ctr =~ s/^\s*(.+)/$1/;
            $ctr =~ tr/ /_/;
            $value =~ s/(.+\s.+)/\"$1\"/;
            $ctrPair{$ctr} = $value;
         }
      }
   }
   else {
      $ctr = $atribName;
      $value = $atribVal;
      if ( $value =~ /^(\d{4})[-\/:]*(\d{2})[-\/:]*(\d{2}).(\d{2}:\d{2}):\d{2}([+-]*\d{0,2}:*\d{0,2})$/o  ) {
         my ($yyyy,$mm,$dd,$tt,$gmt)=($1,$2,$3,$4,$5);
         if ($gmt) {
            $tmpString = $ctr.'_GMT_Difference';
            $tmpString =~ s/^\s*(.+)/$1/;
            $tmpString =~ tr/ /_/;
            $ctrPair{$tmpString} = $gmt;
         }
         if ($tt) {
            $tmpString = $ctr.'_Time';
            $tmpString =~ s/^\s*(.+)/$1/;
            $tmpString =~ tr/ /_/;
            $ctrPair{$tmpString} = $tt;
         }
         $ctr .= '_Date';
         $value = $dd.$MONTH_NAMES{$mm}.$yyyy;
      }
      elsif ( $value =~ /^PT(\d{4})S$/o ) {
         $value = $1;
      }
      $ctr =~ s/^\s*(.+)/$1/;
      $ctr =~ tr/ /_/;
      $value =~ s/(.+\s.+)/\"$1\"/;
      $ctrPair{$ctr} = $value;
   }

   return %ctrPair;
}

################################################################################
# Subroutine name:  invoke_header_data_record_processing()
#
# Description:      Invoke HEADER_DATA_RECORD_PROCESSING subroutine
#
# Arguments:        Block name, header hash ref, data hash ref
#
# Returns:          Object reference
#
sub invoke_header_data_record_processing {
       my ($self, $block_name_ref, $header_ref, $data_ref) = @_;

       if ( $self->{HEADER_DATA_RECORD_PROCESSING} ) {
           my $result = $self->{HEADER_DATA_RECORD_PROCESSING}($block_name_ref, $header_ref, $data_ref);
           return $result;
       }
       return 0;
}

################################################################################
#
# Subroutine name: invoke_configured_record_processing
#
# Description: execute the codes written in EngineConfig.pm
# 	       Maintained for backward compatibility of configurations
#
# Arguments: $sub_address_ref  - This is the reference to the code
#            $ctr_rec_ref      - This is the reference to the unpacked data
#            $op_h_ref         - This is the reference to the hash of pif data
#
# Returns:   $result  - value returned from the executed code
#            -2       - error executing the code
#
################################################################################
sub invoke_configured_record_processing {
    my ($self, $sub_address_ref, $op_h_ref) = @_;

    my $result = -1;
    eval { $result = &$sub_address_ref ($op_h_ref); };
    if ($@) {
        LogMess("invoke_configured_record_processing Error : subroutine failed : '$@' ",2);
        return -2;
    }
    if ($result != 0) {
        LogMess("invoke_configured_record_processing Error : subroutine failed",2);
        return;
    }
}

################################################################################
#
# Subroutine name:  getHeaderStatsData
#
# Description:      Fills out the header statistics information so that it can
#                   be used by the stats processsing. If you have stats switched
#                   on for your module you must fill this function out.
#
# Arguments:        a reference to the statistics object to fill out
#
# Returns:          None
#
################################################################################
sub getHeaderStatsData {
    my ($self,$statsRef) = @_;

    foreach my $k (keys %{$self->{commonInternalHeaderData}}) {
    }
    
    my ($statsName,$counterName);
    
    # the header information is usually derived from the pif file
    # header, being stored in $self->{headerData}
    while (($statsName,$counterName) = each %{$statsRef}) {
        if (exists( $self->{commonInternalHeaderData}->{$counterName})){
            $statsRef->{$statsName} = $self->{commonInternalHeaderData}->{$counterName};
        }
        else {
            LogMess("Warning - could not find entry $counterName " .
                    " in header data for statistics ",3);
            $statsRef->{$statsName} = "";
        }
    }
}

1;  # tells perl that the package is ready to run
